"""Representation of notes and in the vault."""


import copy
import difflib
import re
from pathlib import Path

import rich.repr
import typer
from rich.table import Table

from obsidian_metadata._utils import alerts
from obsidian_metadata._utils.alerts import logger as log
from obsidian_metadata._utils.console import console
from obsidian_metadata.models import (
    Frontmatter,
    InlineMetadata,
    InlineTags,
    InsertLocation,
    MetadataType,
    Patterns,
)

PATTERNS = Patterns()


@rich.repr.auto
class Note:
    """Representation of a note in the vault.

    Args:
        note_path (Path): Path to the note file.

    Attributes:
        note_path (Path): Path to the note file.
        dry_run (bool): Whether to run in dry-run mode.
        file_content (str): Total contents of the note file (frontmatter and content).
        frontmatter (dict): Frontmatter of the note.
        inline_tags (list): List of inline tags in the note.
        inline_metadata (dict): Dictionary of inline metadata in the note.
    """

    def __init__(self, note_path: Path, dry_run: bool = False):
        log.trace(f"Creating Note object for {note_path}")
        self.note_path: Path = Path(note_path)
        self.dry_run: bool = dry_run

        try:
            with self.note_path.open():
                self.file_content: str = self.note_path.read_text()
        except FileNotFoundError as e:
            alerts.error(f"Note {self.note_path} not found. Exiting")
            raise typer.Exit(code=1) from e

        try:
            self.frontmatter: Frontmatter = Frontmatter(self.file_content)
        except AttributeError as e:
            alerts.error(f"Note {self.note_path} has invalid frontmatter.\n{e}")
            raise typer.Exit(code=1) from e

        self.inline_tags: InlineTags = InlineTags(self.file_content)
        self.inline_metadata: InlineMetadata = InlineMetadata(self.file_content)
        self.original_file_content: str = self.file_content

    def __rich_repr__(self) -> rich.repr.Result:  # pragma: no cover
        """Define rich representation of Vault."""
        yield "note_path", self.note_path
        yield "dry_run", self.dry_run
        yield "frontmatter", self.frontmatter
        yield "inline_tags", self.inline_tags
        yield "inline_metadata", self.inline_metadata

    def _delete_inline_metadata(self, key: str, value: str = None) -> None:
        """Delete an inline metadata key/value pair from the text of the note. This method does not remove the key/value from the metadata attribute of the note.

        Args:
            key (str): Key to delete.
            value (str, optional): Value to delete.
        """
        all_results = PATTERNS.find_inline_metadata.findall(self.file_content)
        stripped_null_values = [tuple(filter(None, x)) for x in all_results]

        for _k, _v in stripped_null_values:
            if re.search(key, _k):
                if value is None:
                    _k = re.escape(_k)
                    _v = re.escape(_v)
                    self.sub(rf"\[?{_k}:: ?{_v}]?", "", is_regex=True)
                    return

                if re.search(value, _v):
                    _k = re.escape(_k)
                    _v = re.escape(_v)
                    self.sub(rf"({_k}::) ?{_v}", r"\1", is_regex=True)

    def _rename_inline_metadata(self, key: str, value_1: str, value_2: str = None) -> None:
        """Replace the inline metadata in the note with the current inline metadata object.

        Args:
            key (str): Key to rename.
            value_1 (str): Value to replace OR new key name (if value_2 is None).
            value_2 (str, optional): New value.

        """
        all_results = PATTERNS.find_inline_metadata.findall(self.file_content)
        stripped_null_values = [tuple(filter(None, x)) for x in all_results]

        for _k, _v in stripped_null_values:
            if re.search(key, _k):
                if value_2 is None:
                    if re.search(rf"{key}[^\w\d_-]+", _k):
                        key_text = re.split(r"[^\w\d_-]+$", _k)[0]
                        key_markdown = re.split(r"^[\w\d_-]+", _k)[1]
                        self.sub(
                            rf"{key_text}{key_markdown}::",
                            rf"{value_1}{key_markdown}::",
                        )
                    else:
                        self.sub(f"{_k}::", f"{value_1}::")
                elif re.search(key, _k) and re.search(value_1, _v):
                    _k = re.escape(_k)
                    _v = re.escape(_v)
                    self.sub(f"{_k}:: ?{_v}", f"{_k}:: {value_2}", is_regex=True)

    def add_metadata(  # noqa: C901
        self,
        area: MetadataType,
        key: str = None,
        value: str | list[str] = None,
        location: InsertLocation = None,
    ) -> bool:
        """Add metadata to the note if it does not already exist.

        Args:
            area (MetadataType): Area to add metadata to.
            key (str, optional): Key to add
            location (InsertLocation, optional): Location to add inline metadata and tags.
            value (str, optional): Value to add.

        Returns:
            bool: Whether the metadata was added.
        """
        match area:
            case MetadataType.FRONTMATTER if self.frontmatter.add(key, value):
                self.update_frontmatter()
                return True

            case MetadataType.INLINE:
                if value is None and self.inline_metadata.add(key):
                    line = f"{key}::"
                    self.insert(new_string=line, location=location)
                    return True

                new_values = []
                if isinstance(value, list):
                    new_values = [_v for _v in value if self.inline_metadata.add(key, _v)]
                elif self.inline_metadata.add(key, value):
                    new_values = [value]

                if new_values:
                    for value in new_values:
                        self.insert(new_string=f"{key}:: {value}", location=location)
                    return True

            case MetadataType.TAGS:
                new_values = []
                if isinstance(value, list):
                    new_values = [_v for _v in value if self.inline_tags.add(_v)]
                elif self.inline_tags.add(value):
                    new_values = [value]

                if new_values:
                    for value in new_values:
                        _v = value
                        if _v.startswith("#"):
                            _v = _v[1:]
                        self.insert(new_string=f"#{_v}", location=location)
                    return True

            case _:
                return False

        return False

    def contains_inline_tag(self, tag: str, is_regex: bool = False) -> bool:
        """Check if a note contains the specified inline tag.

        Args:
            tag (str): Tag to check for.
            is_regex (bool, optional): Whether to use regex to match the tag.

        Returns:
            bool: Whether the note has inline tags.
        """
        return self.inline_tags.contains(tag, is_regex=is_regex)

    def contains_metadata(self, key: str, value: str = None, is_regex: bool = False) -> bool:
        """Check if a note has a key or a key-value pair in its metadata.

        Args:
            key (str): Key to check for.
            value (str, optional): Value to check for.
            is_regex (bool, optional): Whether to use regex to match the key/value.

        Returns:
            bool: Whether the note contains the key or key-value pair.
        """
        if value is None:
            if self.frontmatter.contains(key, is_regex=is_regex) or self.inline_metadata.contains(
                key, is_regex=is_regex
            ):
                return True
            return False

        if self.frontmatter.contains(
            key, value, is_regex=is_regex
        ) or self.inline_metadata.contains(key, value, is_regex=is_regex):
            return True

        return False

    def delete_inline_tag(self, tag: str) -> bool:
        """Delete an inline tag from the `inline_tags` attribute AND removes the tag from the text of the note if it exists.

        Args:
            tag (str): Tag to delete.

        Returns:
            bool: Whether the tag was deleted.
        """
        new_list = self.inline_tags.list.copy()

        for _t in new_list:
            if re.search(tag, _t):
                _t = re.escape(_t)
                self.sub(rf"#{_t}([ \|,;:\*\(\)\[\]\\\.\n#&])", r"\1", is_regex=True)
                self.inline_tags.delete(tag)

        if new_list != self.inline_tags.list:
            return True

        return False

    def delete_metadata(
        self, key: str, value: str = None, area: MetadataType = MetadataType.ALL
    ) -> bool:
        """Delete a key or key-value pair from the note's metadata. Regex is supported.

        If no value is provided, will delete an entire key.

        Args:
            key (str): Key to delete.
            value (str, optional): Value to delete.
            area (MetadataType, optional): Area to delete metadata from. Defaults to MetadataType.ALL.

        Returns:
            bool: Whether the key or key-value pair was deleted.
        """
        changed_value: bool = False

        if value is None:
            if (
                area == MetadataType.FRONTMATTER or area == MetadataType.ALL
            ) and self.frontmatter.delete(key):
                self.update_frontmatter()
                changed_value = True
            if (
                area == MetadataType.INLINE or area == MetadataType.ALL
            ) and self.inline_metadata.delete(key):
                self._delete_inline_metadata(key, value)
                changed_value = True
        else:
            if (
                area == MetadataType.FRONTMATTER or area == MetadataType.ALL
            ) and self.frontmatter.delete(key, value):
                self.update_frontmatter()
                changed_value = True
            if (
                area == MetadataType.INLINE or area == MetadataType.ALL
            ) and self.inline_metadata.delete(key, value):
                self._delete_inline_metadata(key, value)
                changed_value = True

        if changed_value:
            return True
        return False

    def has_changes(self) -> bool:
        """Check if the note has been updated.

        Returns:
            bool: Whether the note has been updated.
        """
        if self.frontmatter.has_changes():
            return True

        if self.inline_tags.has_changes():
            return True

        if self.inline_metadata.has_changes():
            return True

        if self.file_content != self.original_file_content:
            return True

        return False

    def insert(
        self,
        new_string: str,
        location: InsertLocation,
        allow_multiple: bool = False,
    ) -> None:
        """Insert a string at the top of a note.

        Args:
            new_string (str): String to insert at the top of the note.
            allow_multiple (bool): Whether to allow inserting the string if it already exists in the note.
            location (InsertLocation): Location to insert the string.
        """
        if not allow_multiple and len(re.findall(re.escape(new_string), self.file_content)) > 0:
            return

        match location:
            case InsertLocation.BOTTOM:
                self.file_content += f"\n{new_string}"
            case InsertLocation.TOP:
                try:
                    top = PATTERNS.frontmatter_block.search(self.file_content).group("frontmatter")
                except AttributeError:
                    top = ""

                if top == "":
                    self.file_content = f"{new_string}\n{self.file_content}"
                else:
                    new_string = f"{top}\n{new_string}"
                    top = re.escape(top)
                    self.sub(top, new_string, is_regex=True)
            case InsertLocation.AFTER_TITLE:
                try:
                    top = PATTERNS.top_with_header.search(self.file_content).group("top")
                except AttributeError:
                    top = ""

                if top == "":
                    self.file_content = f"{new_string}\n{self.file_content}"
                else:
                    new_string = f"{top}\n{new_string}"
                    top = re.escape(top)
                    self.sub(top, new_string, is_regex=True)
            case _:
                raise ValueError(f"Invalid location: {location}")
        pass

    def print_note(self) -> None:
        """Print the note to the console."""
        console.print(self.file_content)

    def print_diff(self) -> None:
        """Print a diff of the note's original state and it's new state."""
        a = self.original_file_content.splitlines()
        b = self.file_content.splitlines()

        diff = difflib.Differ()
        result = list(diff.compare(a, b))
        table = Table(title=f"\nDiff of {self.note_path.name}", show_header=False, min_width=50)

        for line in result:
            if line.startswith("+"):
                table.add_row(line, style="green")
            elif line.startswith("-"):
                table.add_row(line, style="red")

        console.print(table)

    def rename_inline_tag(self, tag_1: str, tag_2: str) -> bool:
        """Rename an inline tag from the note ONLY if it's not in the metadata as well.

        Args:
            tag_1 (str): Tag to rename.
            tag_2 (str): New tag name.

        Returns:
            bool: Whether the tag was renamed.
        """
        if tag_1 in self.inline_tags.list:
            self.sub(
                rf"#{tag_1}([ \|,;:\*\(\)\[\]\\\.\n#&])",
                rf"#{tag_2}\1",
                is_regex=True,
            )
            self.inline_tags.rename(tag_1, tag_2)
            return True
        return False

    def rename_metadata(self, key: str, value_1: str, value_2: str = None) -> bool:
        """Rename a key or key-value pair in the note's metadata.

        If no value is provided, will rename an entire key.

        Args:
            key (str): Key to rename.
            value_1 (str): Value to rename or new name of key if no value_2 is provided.
            value_2 (str, optional): New value.

        Returns:
            bool: Whether the note was updated.
        """
        changed_value: bool = False
        if value_2 is None:
            if self.frontmatter.rename(key, value_1):
                self.update_frontmatter()
                changed_value = True
            if self.inline_metadata.rename(key, value_1):
                self._rename_inline_metadata(key, value_1)
                changed_value = True
        else:
            if self.frontmatter.rename(key, value_1, value_2):
                self.update_frontmatter()
                changed_value = True
            if self.inline_metadata.rename(key, value_1, value_2):
                self._rename_inline_metadata(key, value_1, value_2)
                changed_value = True

        if changed_value:
            return True

        return False

    def sub(self, pattern: str, replacement: str, is_regex: bool = False) -> None:
        """Substitutes text within the note.

        Args:
            pattern (str): The pattern to replace (plain text or regular expression).
            replacement (str): What to replace the pattern with.
            is_regex (bool): Whether the pattern is a regex pattern or plain text.
        """
        if not is_regex:
            pattern = re.escape(pattern)

        self.file_content = re.sub(pattern, replacement, self.file_content, re.MULTILINE)

    def transpose_metadata(  # noqa: C901, PLR0912, PLR0911, PLR0913
        self,
        begin: MetadataType,
        end: MetadataType,
        key: str = None,
        value: str | list[str] = None,
        location: InsertLocation = InsertLocation.BOTTOM,
    ) -> bool:
        """Transpose metadata from one type to another.

        Args:
            begin (MetadataType): The type of metadata to transpose from.
            end (MetadataType): The type of metadata to transpose to.
            key (str, optional): The key to transpose. Defaults to None.
            value (str | list[str], optional): The value to transpose. Defaults to None.

        Returns:
            bool: Whether the note was updated.
        """
        if (begin == MetadataType.FRONTMATTER or begin == MetadataType.INLINE) and (
            end == MetadataType.FRONTMATTER or end == MetadataType.INLINE
        ):
            if begin == MetadataType.FRONTMATTER:
                begin_dict = self.frontmatter.dict
            else:
                begin_dict = self.inline_metadata.dict

            if begin_dict == {}:
                return False

            if key is None:  # Transpose all metadata when no key is provided
                for _key, _value in begin_dict.items():
                    self.add_metadata(key=_key, value=_value, area=end, location=location)
                    self.delete_metadata(key=_key, area=begin)
                return True

            has_changes = False
            temp_dict = copy.deepcopy(begin_dict)
            for k, v in begin_dict.items():
                if key == k:
                    if value is None:
                        self.add_metadata(key=k, value=v, area=end, location=location)
                        self.delete_metadata(key=k, area=begin)
                        return True

                    if value == v:
                        self.add_metadata(key=k, value=v, area=end, location=location)
                        self.delete_metadata(key=k, area=begin)
                        return True

                    if isinstance(value, str):
                        if value in v:
                            self.add_metadata(key=k, value=value, area=end, location=location)
                            self.delete_metadata(key=k, value=value, area=begin)
                            return True

                        return False

                    if isinstance(value, list):
                        for value_item in value:
                            if value_item in v:
                                self.add_metadata(
                                    key=k, value=value_item, area=end, location=location
                                )
                                self.delete_metadata(key=k, value=value_item, area=begin)
                                temp_dict[k].remove(value_item)
                                has_changes = True

                        if temp_dict[k] == []:
                            self.delete_metadata(key=k, area=begin)

                    return bool(has_changes)

        if begin == MetadataType.TAGS:
            # TODO: Implement transposing to and from tags
            pass

        return False

    def update_frontmatter(self, sort_keys: bool = False) -> None:
        """Replace the frontmatter in the note with the current frontmatter object."""
        try:
            current_frontmatter = PATTERNS.frontmatter_block.search(self.file_content).group(
                "frontmatter"
            )
        except AttributeError:
            current_frontmatter = None

        if current_frontmatter is None and self.frontmatter.dict == {}:
            return

        new_frontmatter = self.frontmatter.to_yaml(sort_keys=sort_keys)
        new_frontmatter = "" if self.frontmatter.dict == {} else f"---\n{new_frontmatter}---\n"

        if current_frontmatter is None:
            self.file_content = new_frontmatter + self.file_content
            return

        current_frontmatter = f"{re.escape(current_frontmatter)}\n?"
        self.sub(current_frontmatter, new_frontmatter, is_regex=True)

    def write(self, path: Path = None) -> None:
        """Write the note's content to disk.

        Args:
            path (Path): Path to write the note to. Defaults to the note's path.

        Raises:
            typer.Exit: If the note's path is not found.
        """
        p = self.note_path if path is None else path
        if self.dry_run:
            log.trace(f"DRY RUN: Writing note {p} to disk")
            return

        try:
            with open(p, "w") as f:
                log.trace(f"Writing note {p} to disk")
                f.write(self.file_content)
        except FileNotFoundError as e:
            alerts.error(f"Note {p} not found. Exiting")
            raise typer.Exit(code=1) from e



"""Work with metadata items."""

import copy
import re
from io import StringIO

from rich.columns import Columns
from rich.table import Table
from ruamel.yaml import YAML

from obsidian_metadata._utils import (
    clean_dictionary,
    dict_contains,
    dict_values_to_lists_strings,
    merge_dictionaries,
    remove_markdown_sections,
)
from obsidian_metadata._utils.console import console
from obsidian_metadata.models import Patterns  # isort: ignore
from obsidian_metadata.models.enums import MetadataType

PATTERNS = Patterns()
INLINE_TAG_KEY: str = "inline_tag"


class VaultMetadata:
    """Representation of all Metadata in the Vault."""

    def __init__(self) -> None:
        self.dict: dict[str, list[str]] = {}
        self.frontmatter: dict[str, list[str]] = {}
        self.inline_metadata: dict[str, list[str]] = {}
        self.tags: list[str] = []

    def __repr__(self) -> str:
        """Representation of all metadata."""
        return str(self.dict)

    def index_metadata(
        self, area: MetadataType, metadata: dict[str, list[str]] | list[str]
    ) -> None:
        """Index pre-existing metadata in the vault. Takes a dictionary as input and merges it with the existing metadata.  Does not overwrite existing keys.

        Args:
            area (MetadataType): Type of metadata.
            metadata (dict): Metadata to add.
        """
        if isinstance(metadata, dict):
            new_metadata = clean_dictionary(metadata)
            self.dict = merge_dictionaries(self.dict, new_metadata)

        if area == MetadataType.FRONTMATTER:
            self.frontmatter = merge_dictionaries(self.frontmatter, new_metadata)

        if area == MetadataType.INLINE:
            self.inline_metadata = merge_dictionaries(self.inline_metadata, new_metadata)

        if area == MetadataType.TAGS and isinstance(metadata, list):
            self.tags.extend(metadata)
            self.tags = sorted({s.strip("#") for s in self.tags})

    def contains(  # noqa: PLR0911
        self, area: MetadataType, key: str = None, value: str = None, is_regex: bool = False
    ) -> bool:
        """Check if a key and/or a value exists in the metadata.

        Args:
            area (MetadataType): Type of metadata to check.
            key (str, optional): Key to check.
            value (str, optional): Value to check.
            is_regex (bool, optional): Use regex to check. Defaults to False.

        Returns:
            bool: True if the key exists.

        Raises:
            ValueError: Key must be provided when checking for a key's existence.
            ValueError: Value must be provided when checking for a tag's existence.
        """
        if area != MetadataType.TAGS and key is None:
            raise ValueError("Key must be provided when checking for a key's existence.")

        match area:
            case MetadataType.ALL:
                if dict_contains(self.dict, key, value, is_regex):
                    return True
                if key is None and value is not None:
                    if is_regex:
                        return any(re.search(value, tag) for tag in self.tags)
                    return value in self.tags

            case MetadataType.FRONTMATTER:
                return dict_contains(self.frontmatter, key, value, is_regex)
            case MetadataType.INLINE:
                return dict_contains(self.inline_metadata, key, value, is_regex)
            case MetadataType.KEYS:
                return dict_contains(self.dict, key, value, is_regex)
            case MetadataType.TAGS:
                if value is None:
                    raise ValueError("Value must be provided when checking for a tag's existence.")
                if is_regex:
                    return any(re.search(value, tag) for tag in self.tags)
                return value in self.tags

        return False

    def delete(self, key: str, value_to_delete: str = None) -> bool:
        """Delete a key or a key's value from the metadata. Regex is supported to allow deleting more than one key or value.

        Args:
            key (str): Key to check.
            value_to_delete (str, optional): Value to delete.

        Returns:
            bool: True if a value was deleted
        """
        new_dict = copy.deepcopy(self.dict)

        if value_to_delete is None:
            for _k in list(new_dict):
                if re.search(key, _k):
                    del new_dict[_k]
        else:
            for _k, _v in new_dict.items():
                if re.search(key, _k):
                    new_values = [x for x in _v if not re.search(value_to_delete, x)]
                    new_dict[_k] = sorted(new_values)

        if new_dict != self.dict:
            self.dict = dict(new_dict)
            return True

        return False

    def print_metadata(self, area: MetadataType) -> None:
        """Print metadata to the terminal.

        Args:
            area (MetadataType): Type of metadata to print
        """
        dict_to_print: dict[str, list[str]] = None
        list_to_print: list[str] = None
        match area:
            case MetadataType.INLINE:
                dict_to_print = self.inline_metadata.copy()
                header = "All inline metadata"
            case MetadataType.FRONTMATTER:
                dict_to_print = self.frontmatter.copy()
                header = "All frontmatter"
            case MetadataType.TAGS:
                list_to_print = []
                for tag in self.tags:
                    list_to_print.append(f"#{tag}")
                header = "All inline tags"
            case MetadataType.KEYS:
                list_to_print = sorted(self.dict.keys())
                header = "All Keys"
            case MetadataType.ALL:
                dict_to_print = self.dict.copy()
                list_to_print = []
                for tag in self.tags:
                    list_to_print.append(f"#{tag}")
                header = "All metadata"

        if dict_to_print is not None:
            table = Table(title=header, show_footer=False, show_lines=True)
            table.add_column("Keys")
            table.add_column("Values")
            for key, value in sorted(dict_to_print.items()):
                values: str | dict[str, list[str]] = (
                    "\n".join(sorted(value)) if isinstance(value, list) else value
                )
                table.add_row(f"[bold]{key}[/]", str(values))
            console.print(table)

        if list_to_print is not None:
            columns = Columns(
                sorted(list_to_print),
                equal=True,
                expand=True,
                title=header if area != MetadataType.ALL else "All inline tags",
            )
            console.print(columns)

    def rename(self, key: str, value_1: str, value_2: str = None) -> bool:
        """Replace a value in the frontmatter.

        Args:
            key (str): Key to check.
            value_1 (str): `With value_2` this is the value to rename. If `value_2` is None this is the renamed key
            value_2 (str, Optional): New value.
            bypass_check (bool, optional): Bypass the check if the key exists. Defaults to False.

        Returns:
            bool: True if a value was renamed
        """
        if value_2 is None:
            if key in self.dict and value_1 not in self.dict:
                self.dict[value_1] = self.dict.pop(key)
                return True
            return False

        if key in self.dict and value_1 in self.dict[key]:
            self.dict[key] = sorted({value_2 if x == value_1 else x for x in self.dict[key]})
            return True

        return False


class Frontmatter:
    """Representation of frontmatter metadata."""

    def __init__(self, file_content: str):
        self.dict: dict[str, list[str]] = self._grab_note_frontmatter(file_content)
        self.dict_original: dict[str, list[str]] = copy.deepcopy(self.dict)

    def __repr__(self) -> str:  # pragma: no cover
        """Representation of the frontmatter.

        Returns:
            str: frontmatter
        """
        return f"Frontmatter(frontmatter={self.dict})"

    def _grab_note_frontmatter(self, file_content: str) -> dict:
        """Grab metadata from a note.

        Args:
            note_path (Path): Path to the note file.

        Returns:
            dict: Metadata from the note.
        """
        try:
            frontmatter_block: str = PATTERNS.frontmatt_block_strip_separators.search(
                file_content
            ).group("frontmatter")
        except AttributeError:
            return {}

        yaml = YAML(typ="safe")
        yaml.allow_unicode = False
        try:
            frontmatter: dict = yaml.load(frontmatter_block)
        except Exception as e:  # noqa: BLE001
            raise AttributeError(e) from e

        for k in frontmatter:
            if frontmatter[k] is None:
                frontmatter[k] = []

        return dict_values_to_lists_strings(frontmatter, strip_null_values=True)

    def add(self, key: str, value: str | list[str] = None) -> bool:  # noqa: PLR0911
        """Add a key and value to the frontmatter.

        Args:
            key (str): Key to add.
            value (str, optional): Value to add.

        Returns:
            bool: True if the metadata was added
        """
        if value is None:
            if key not in self.dict:
                self.dict[key] = []
                return True
            return False

        if key not in self.dict:
            if isinstance(value, list):
                self.dict[key] = value
                return True

            self.dict[key] = [value]
            return True

        if key in self.dict and value not in self.dict[key]:
            if isinstance(value, list):
                self.dict[key].extend(value)
                return True

            self.dict[key].append(value)
            return True

        return False

    def contains(self, key: str, value: str = None, is_regex: bool = False) -> bool:
        """Check if a key or value exists in the metadata.

        Args:
            key (str): Key to check.
            value (str, optional): Value to check.
            is_regex (bool, optional): Use regex to check. Defaults to False.

        Returns:
            bool: True if the key exists.
        """
        return dict_contains(self.dict, key, value, is_regex)

    def delete(self, key: str, value_to_delete: str = None) -> bool:
        """Delete a value or key in the frontmatter.  Regex is supported to allow deleting more than one key or value.

        Args:
            key (str): If no value, key to delete. If value, key containing the value.
            value_to_delete (str, optional): Value to delete.

        Returns:
            bool: True if a value was deleted
        """
        new_dict = dict(self.dict)

        if value_to_delete is None:
            for _k in list(new_dict):
                if re.search(key, _k):
                    del new_dict[_k]
        else:
            for _k, _v in new_dict.items():
                if re.search(key, _k):
                    new_values = [x for x in _v if not re.search(value_to_delete, x)]
                    new_dict[_k] = sorted(new_values)

        if new_dict != self.dict:
            self.dict = dict(new_dict)
            return True

        return False

    def has_changes(self) -> bool:
        """Check if the frontmatter has changes.

        Returns:
            bool: True if the frontmatter has changes.
        """
        return self.dict != self.dict_original

    def rename(self, key: str, value_1: str, value_2: str = None) -> bool:
        """Replace a value in the frontmatter.

        Args:
            key (str): Key to check.
            value_1 (str): `With value_2` this is the value to rename. If `value_2` is None this is the renamed key
            value_2 (str, Optional): New value.

        Returns:
            bool: True if a value was renamed
        """
        if value_2 is None:
            if key in self.dict and value_1 not in self.dict:
                self.dict[value_1] = self.dict.pop(key)
                return True
            return False

        if key in self.dict and value_1 in self.dict[key]:
            self.dict[key] = sorted({value_2 if x == value_1 else x for x in self.dict[key]})
            return True

        return False

    def to_yaml(self, sort_keys: bool = False) -> str:
        """Return the frontmatter as a YAML string.

        Returns:
            str: Frontmatter as a YAML string.
            sort_keys (bool, optional): Sort the keys. Defaults to False.
        """
        dict_to_dump = copy.deepcopy(self.dict)
        for k in dict_to_dump:
            if dict_to_dump[k] == []:
                dict_to_dump[k] = None
            if isinstance(dict_to_dump[k], list) and len(dict_to_dump[k]) == 1:
                new_val = dict_to_dump[k][0]
                dict_to_dump[k] = new_val  # type: ignore [assignment]

        # Converting stream to string from https://stackoverflow.com/questions/47614862/best-way-to-use-ruamel-yaml-to-dump-yaml-to-string-not-to-stream/63179923#63179923

        if sort_keys:
            dict_to_dump = dict(sorted(dict_to_dump.items()))

        yaml = YAML()
        yaml.indent(mapping=2, sequence=4, offset=2)
        string_stream = StringIO()
        yaml.dump(dict_to_dump, string_stream)
        yaml_value = string_stream.getvalue()
        string_stream.close()
        return yaml_value


class InlineMetadata:
    """Representation of inline metadata in the form of `key:: value`."""

    def __init__(self, file_content: str):
        self.dict: dict[str, list[str]] = self.grab_inline_metadata(file_content)
        self.dict_original: dict[str, list[str]] = copy.deepcopy(self.dict)

    def __repr__(self) -> str:  # pragma: no cover
        """Representation of inline metadata.

        Returns:
            str: inline metadata
        """
        return f"InlineMetadata(inline_metadata={self.dict})"

    def add(self, key: str, value: str | list[str] = None) -> bool:  # noqa: PLR0911
        """Add a key and value to the inline metadata.

        Args:
            key (str): Key to add.
            value (str, optional): Value to add.

        Returns:
            bool: True if the metadata was added
        """
        if value is None:
            if key not in self.dict:
                self.dict[key] = []
                return True
            return False

        if key not in self.dict:
            if isinstance(value, list):
                self.dict[key] = value
                return True

            self.dict[key] = [value]
            return True

        if key in self.dict and value not in self.dict[key]:
            if isinstance(value, list):
                self.dict[key].extend(value)
                return True

            self.dict[key].append(value)
            return True

        return False

    def contains(self, key: str, value: str = None, is_regex: bool = False) -> bool:
        """Check if a key or value exists in the inline metadata.

        Args:
            key (str): Key to check.
            value (str, Optional): Value to check.
            is_regex (bool, optional): If True, key and value are treated as regex. Defaults to False.

        Returns:
            bool: True if the key exists.
        """
        return dict_contains(self.dict, key, value, is_regex)

    def delete(self, key: str, value_to_delete: str = None) -> bool:
        """Delete a value or key in the inline metadata. Regex is supported to allow deleting more than one key or value.

        Args:
            key (str): If no value, key to delete. If value, key containing the value.
            value_to_delete (str, optional): Value to delete.

        Returns:
            bool: True if a value was deleted
        """
        new_dict = dict(self.dict)

        if value_to_delete is None:
            for _k in list(new_dict):
                if re.search(key, _k):
                    del new_dict[_k]
        else:
            for _k, _v in new_dict.items():
                if re.search(key, _k):
                    new_values = [x for x in _v if not re.search(value_to_delete, x)]
                    new_dict[_k] = sorted(new_values)

        if new_dict != self.dict:
            self.dict = dict(new_dict)
            return True

        return False

    def grab_inline_metadata(self, file_content: str) -> dict[str, list[str]]:
        """Grab inline metadata from a note.

        Returns:
            dict[str, str]: Inline metadata from the note.
        """
        content = remove_markdown_sections(
            file_content,
            strip_codeblocks=True,
            strip_inlinecode=True,
            strip_frontmatter=True,
        )
        all_results = PATTERNS.find_inline_metadata.findall(content)
        stripped_null_values = [tuple(filter(None, x)) for x in all_results]

        inline_metadata: dict[str, list[str]] = {}
        for k, v in stripped_null_values:
            if k in inline_metadata:
                inline_metadata[k].append(str(v))
            else:
                inline_metadata[k] = [str(v)]

        return clean_dictionary(inline_metadata)

    def has_changes(self) -> bool:
        """Check if the metadata has changes.

        Returns:
            bool: True if the metadata has changes.
        """
        return self.dict != self.dict_original

    def rename(self, key: str, value_1: str, value_2: str = None) -> bool:
        """Replace a value in the inline metadata.

        Args:
            key (str): Key to check.
            value_1 (str): `With value_2` this is the value to rename. If `value_2` is None this is the renamed key
            value_2 (str, Optional): New value.

        Returns:
            bool: True if a value was renamed
        """
        if value_2 is None:
            if key in self.dict and value_1 not in self.dict:
                self.dict[value_1] = self.dict.pop(key)
                return True
            return False

        if key in self.dict and value_1 in self.dict[key]:
            self.dict[key] = sorted({value_2 if x == value_1 else x for x in self.dict[key]})
            return True

        return False


class InlineTags:
    """Representation of inline tags."""

    def __init__(self, file_content: str):
        self.metadata_key = INLINE_TAG_KEY
        self.list: list[str] = self._grab_inline_tags(file_content)
        self.list_original: list[str] = self.list.copy()

    def __repr__(self) -> str:  # pragma: no cover
        """Representation of the inline tags.

        Returns:
            str: inline tags
        """
        return f"InlineTags(tags={self.list})"

    def _grab_inline_tags(self, file_content: str) -> list[str]:
        """Grab inline tags from a note.

        Args:
            file_content (str): Total contents of the note file (frontmatter and content).

        Returns:
            list[str]: Inline tags from the note.
        """
        return sorted(
            PATTERNS.find_inline_tags.findall(
                remove_markdown_sections(
                    file_content,
                    strip_codeblocks=True,
                    strip_inlinecode=True,
                )
            )
        )

    def add(self, new_tag: str | list[str]) -> bool:
        """Add a new inline tag.

        Args:
            new_tag (str): Tag to add.

        Returns:
            bool: True if a tag was added.
        """
        if isinstance(new_tag, list):
            for _tag in new_tag:
                if _tag.startswith("#"):
                    _tag = _tag[1:]
                if _tag in self.list:
                    return False
                new_list = self.list.copy()
                new_list.append(_tag)
                self.list = sorted(new_list)
                return True
        else:
            if new_tag.startswith("#"):
                new_tag = new_tag[1:]
            if new_tag in self.list:
                return False
            new_list = self.list.copy()
            new_list.append(new_tag)
            self.list = sorted(new_list)
            return True

        return False

    def contains(self, tag: str, is_regex: bool = False) -> bool:
        """Check if a tag exists in the metadata.

        Args:
            tag (str): Tag to check.
            is_regex (bool, optional): If True, tag is treated as regex. Defaults to False.

        Returns:
            bool: True if the tag exists.
        """
        if is_regex is True:
            return any(re.search(tag, _t) for _t in self.list)

        if tag in self.list:
            return True

        return False

    def delete(self, tag_to_delete: str) -> bool:
        """Delete a specified inline tag. Regex is supported to allow deleting more than one tag.

        Args:
            tag_to_delete (str, optional): Value to delete.

        Returns:
            bool: True if a value was deleted
        """
        new_list = sorted([x for x in self.list if re.search(tag_to_delete, x) is None])

        if new_list != self.list:
            self.list = new_list
            return True
        return False

    def has_changes(self) -> bool:
        """Check if the metadata has changes.

        Returns:
            bool: True if the metadata has changes.
        """
        return self.list != self.list_original

    def rename(self, old_tag: str, new_tag: str) -> bool:
        """Replace an inline tag with another string.

        Args:
            old_tag (str): `With value_2` this is the value to rename. If `value_2` is None this is the renamed key
            new_tag (str, Optional): New value.

        Returns:
            bool: True if a value was renamed
        """
        if old_tag in self.list:
            self.list = sorted([new_tag if i == old_tag else i for i in self.list])
            return True
        return False



"""Questions for the cli."""


from pathlib import Path
from typing import Any

import questionary
from rich import box
from rich.table import Table

from obsidian_metadata._config import VaultConfig
from obsidian_metadata._utils import alerts
from obsidian_metadata._utils.console import console
from obsidian_metadata.models import Vault, VaultFilter
from obsidian_metadata.models.enums import MetadataType
from obsidian_metadata.models.questions import Questions


class Application:
    """Questions for use in the cli.

    Contains methods which ask a series of questions to the user and return a dictionary with their answers.

    More info: https://questionary.readthedocs.io/en/stable/pages/advanced.html#create-questions-from-dictionaries
    """

    def __init__(self, config: VaultConfig, dry_run: bool) -> None:
        self.config = config
        self.dry_run = dry_run
        self.questions = Questions()
        self.filters: list[VaultFilter] = []

    def _load_vault(self) -> None:
        """Load the vault."""
        if len(self.filters) == 0:
            self.vault: Vault = Vault(config=self.config, dry_run=self.dry_run)
        else:
            self.vault = Vault(config=self.config, dry_run=self.dry_run, filters=self.filters)

        alerts.success(
            f"Loaded {len(self.vault.notes_in_scope)} notes from {len(self.vault.all_notes)} total notes"
        )
        self.questions = Questions(vault=self.vault)

    def application_main(self) -> None:
        """Questions for the main application."""
        self._load_vault()

        while True:
            self.vault.info()

            match self.questions.ask_application_main():
                case "vault_actions":
                    self.application_vault()
                case "inspect_metadata":
                    self.application_inspect_metadata()
                case "filter_notes":
                    self.application_filter()
                case "add_metadata":
                    self.application_add_metadata()
                case "rename_metadata":
                    self.application_rename_metadata()
                case "delete_metadata":
                    self.application_delete_metadata()
                case "transpose_metadata":
                    self.application_transpose_metadata()
                case "review_changes":
                    self.review_changes()
                case "commit_changes":
                    self.commit_changes()
                case _:
                    break

        console.print("Done!")
        return

    def application_add_metadata(self) -> None:
        """Add metadata."""
        alerts.usage(
            "Add new metadata to your vault. Currently only supports adding to the frontmatter of a note."
        )

        area = self.questions.ask_area()
        match area:
            case MetadataType.FRONTMATTER | MetadataType.INLINE:
                key = self.questions.ask_new_key(question="Enter the key for the new metadata")
                if key is None:  # pragma: no cover
                    return

                value = self.questions.ask_new_value(
                    question="Enter the value for the new metadata"
                )
                if value is None:  # pragma: no cover
                    return

                num_changed = self.vault.add_metadata(
                    area=area, key=key, value=value, location=self.vault.insert_location
                )
                if num_changed == 0:  # pragma: no cover
                    alerts.warning("No notes were changed")
                    return

                alerts.success(f"Added metadata to {num_changed} notes")

            case MetadataType.TAGS:
                tag = self.questions.ask_new_tag()
                if tag is None:  # pragma: no cover
                    return

                num_changed = self.vault.add_metadata(
                    area=area, value=tag, location=self.vault.insert_location
                )

                if num_changed == 0:  # pragma: no cover
                    alerts.warning("No notes were changed")
                    return

                alerts.success(f"Added metadata to {num_changed} notes")
            case _:  # pragma: no cover
                return

    def application_delete_metadata(self) -> None:
        """Delete metadata."""
        alerts.usage("Delete either a key and all associated values, or a specific value.")

        choices = [
            {"name": "Delete key", "value": "delete_key"},
            {"name": "Delete value", "value": "delete_value"},
            {"name": "Delete inline tag", "value": "delete_inline_tag"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]
        match self.questions.ask_selection(
            choices=choices, question="Select a metadata type to delete"
        ):
            case "delete_key":
                self.delete_key()
            case "delete_value":
                self.delete_value()
            case "delete_inline_tag":
                self.delete_inline_tag()
            case _:  # pragma: no cover
                return

    def application_rename_metadata(self) -> None:
        """Rename metadata."""
        alerts.usage("Select the type of metadata to rename.")

        choices = [
            {"name": "Rename key", "value": "rename_key"},
            {"name": "Rename value", "value": "rename_value"},
            {"name": "Rename inline tag", "value": "rename_inline_tag"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]
        match self.questions.ask_selection(
            choices=choices, question="Select a metadata type to rename"
        ):
            case "rename_key":
                self.rename_key()
            case "rename_value":
                self.rename_value()
            case "rename_inline_tag":
                self.rename_inline_tag()
            case _:  # pragma: no cover
                return

    def application_filter(self) -> None:  # noqa: C901,PLR0911,PLR0912
        """Filter notes."""
        alerts.usage("Limit the scope of notes to be processed with one or more filters.")

        choices = [
            {"name": "Apply new regex path filter", "value": "apply_path_filter"},
            {"name": "Apply new metadata filter", "value": "apply_metadata_filter"},
            {"name": "Apply new in-text tag filter", "value": "apply_tag_filter"},
            {"name": "List and clear filters", "value": "list_filters"},
            {"name": "List notes in scope", "value": "list_notes"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]
        while True:
            match self.questions.ask_selection(choices=choices, question="Select an action"):
                case "apply_path_filter":
                    path = self.questions.ask_filter_path()
                    if path is None or path == "":  # pragma: no cover
                        return

                    self.filters.append(VaultFilter(path_filter=path))
                    self._load_vault()

                case "apply_metadata_filter":
                    key = self.questions.ask_existing_key()
                    if key is None:  # pragma: no cover
                        return

                    questions2 = Questions(vault=self.vault, key=key)
                    value = questions2.ask_existing_value(
                        question="Enter the value for the metadata filter",
                    )
                    if value is None:  # pragma: no cover
                        return
                    if value == "":
                        self.filters.append(VaultFilter(key_filter=key))
                    else:
                        self.filters.append(VaultFilter(key_filter=key, value_filter=value))
                    self._load_vault()

                case "apply_tag_filter":
                    tag = self.questions.ask_existing_inline_tag()
                    if tag is None or tag == "":
                        return

                    self.filters.append(VaultFilter(tag_filter=tag))
                    self._load_vault()

                case "list_filters":
                    if len(self.filters) == 0:
                        alerts.notice("No filters have been applied")
                        return

                    console.print("")
                    table = Table(
                        "Opt",
                        "Filter",
                        "Type",
                        title="Current Filters",
                        show_header=False,
                        box=box.HORIZONTALS,
                    )
                    for _n, _filter in enumerate(self.filters, start=1):
                        if _filter.path_filter is not None:
                            table.add_row(
                                str(_n),
                                f"Path regex: [tan bold]{_filter.path_filter}",
                                end_section=bool(_n == len(self.filters)),
                            )
                        elif _filter.tag_filter is not None:
                            table.add_row(
                                str(_n),
                                f"Tag filter: [tan bold]{_filter.tag_filter}",
                                end_section=bool(_n == len(self.filters)),
                            )
                        elif _filter.key_filter is not None and _filter.value_filter is None:
                            table.add_row(
                                str(_n),
                                f"Key filter: [tan bold]{_filter.key_filter}",
                                end_section=bool(_n == len(self.filters)),
                            )
                        elif _filter.key_filter is not None and _filter.value_filter is not None:
                            table.add_row(
                                str(_n),
                                f"Key/Value : [tan bold]{_filter.key_filter}={_filter.value_filter}",
                                end_section=bool(_n == len(self.filters)),
                            )
                    table.add_row(f"{len(self.filters) + 1}", "Clear All")
                    table.add_row(f"{len(self.filters) + 2}", "Return to Main Menu")
                    console.print(table)

                    num = self.questions.ask_number(
                        question="Enter the number of the filter to clear"
                    )
                    if num is None:
                        return
                    if int(num) <= len(self.filters):
                        self.filters.pop(int(num) - 1)
                        self._load_vault()
                        return
                    if int(num) == len(self.filters) + 1:
                        self.filters = []
                        self._load_vault()
                        return

                case "list_notes":
                    self.vault.list_editable_notes()

                case _:
                    return

    def application_inspect_metadata(self) -> None:
        """View metadata."""
        alerts.usage(
            "Inspect the metadata in your vault. Note, uncommitted changes will be reflected in these reports"
        )

        choices = [
            {"name": "View all metadata", "value": "all_metadata"},
            {"name": "View all frontmatter", "value": "all_frontmatter"},
            {"name": "View all inline_metadata", "value": "all_inline"},
            {"name": "View all keys", "value": "all_keys"},
            {"name": "View all inline tags", "value": "all_tags"},
            questionary.Separator(),
            {"name": "Write all metadata to CSV", "value": "export_csv"},
            {"name": "Write all metadata to JSON file", "value": "export_json"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]
        while True:
            match self.questions.ask_selection(choices=choices, question="Select a vault action"):
                case "all_metadata":
                    console.print("")
                    self.vault.metadata.print_metadata(area=MetadataType.ALL)
                    console.print("")
                case "all_frontmatter":
                    console.print("")
                    self.vault.metadata.print_metadata(area=MetadataType.FRONTMATTER)
                    console.print("")
                case "all_inline":
                    console.print("")
                    self.vault.metadata.print_metadata(area=MetadataType.INLINE)
                    console.print("")
                case "all_keys":
                    console.print("")
                    self.vault.metadata.print_metadata(area=MetadataType.KEYS)
                    console.print("")
                case "all_tags":
                    console.print("")
                    self.vault.metadata.print_metadata(area=MetadataType.TAGS)
                    console.print("")
                case "export_csv":
                    path = self.questions.ask_path(question="Enter a path for the CSV file")
                    if path is None:
                        return
                    self.vault.export_metadata(path=path, export_format="csv")
                    alerts.success(f"Metadata written to {path}")
                case "export_json":
                    path = self.questions.ask_path(question="Enter a path for the JSON file")
                    if path is None:
                        return
                    self.vault.export_metadata(path=path, export_format="json")
                    alerts.success(f"Metadata written to {path}")
                case _:
                    return

    def application_transpose_metadata(self) -> None:
        """Transpose metadata."""
        alerts.usage("Transpose metadata from frontmatter to inline or vice versa.")

        choices = [
            {"name": "Transpose frontmatter to inline", "value": "frontmatter_to_inline"},
            {"name": "Transpose inline to frontmatter", "value": "inline_to_frontmatter"},
        ]
        match self.questions.ask_selection(
            choices=choices, question="Select metadata to transpose"
        ):
            case "frontmatter_to_inline":
                self.transpose_metadata(begin=MetadataType.FRONTMATTER, end=MetadataType.INLINE)
            case "inline_to_frontmatter":
                self.transpose_metadata(begin=MetadataType.INLINE, end=MetadataType.FRONTMATTER)
            case _:  # pragma: no cover
                return

    def application_vault(self) -> None:
        """Vault actions."""
        alerts.usage("Create or delete a backup of your vault.")

        choices = [
            {"name": "Backup vault", "value": "backup_vault"},
            {"name": "Delete vault backup", "value": "delete_backup"},
            questionary.Separator(),
            {"name": "Back", "value": "back"},
        ]

        while True:
            match self.questions.ask_selection(choices=choices, question="Select a vault action"):
                case "backup_vault":
                    self.vault.backup()
                case "delete_backup":
                    self.vault.delete_backup()
                case _:
                    return

    def commit_changes(self) -> bool:
        """Write all changes to disk.

        Returns:
            True if changes were committed, False otherwise.
        """
        changed_notes = self.vault.get_changed_notes()

        if len(changed_notes) == 0:
            console.print("\n")
            alerts.notice("No changes to commit.\n")
            return False

        backup = questionary.confirm("Create backup before committing changes").ask()
        if backup is None:
            return False
        if backup:
            self.vault.backup()

        if questionary.confirm(f"Commit {len(changed_notes)} changed files to disk?").ask():
            self.vault.commit_changes()

        if not self.dry_run:
            alerts.success(f"{len(changed_notes)} changes committed to disk. Exiting")
            return True

        return True

    def delete_inline_tag(self) -> None:
        """Delete an inline tag."""
        tag = self.questions.ask_existing_inline_tag(question="Which tag would you like to delete?")

        num_changed = self.vault.delete_inline_tag(tag)
        if num_changed == 0:
            alerts.warning("No notes were changed")
            return

        alerts.success(f"Deleted inline tag: {tag} in {num_changed} notes")
        return

    def delete_key(self) -> None:
        """Delete a key from the vault."""
        key_to_delete = self.questions.ask_existing_keys_regex(
            question="Regex for the key(s) you'd like to delete?"
        )
        if key_to_delete is None:  # pragma: no cover
            return

        num_changed = self.vault.delete_metadata(key_to_delete)
        if num_changed == 0:
            alerts.warning(f"No notes found with a key matching: [reverse]{key_to_delete}[/]")
            return

        alerts.success(
            f"Deleted keys matching: [reverse]{key_to_delete}[/] from {num_changed} notes"
        )

        return

    def delete_value(self) -> None:
        """Delete a value from the vault."""
        key = self.questions.ask_existing_key(question="Which key contains the value to delete?")
        if key is None:  # pragma: no cover
            return

        questions2 = Questions(vault=self.vault, key=key)
        value = questions2.ask_existing_value_regex(question="Regex for the value to delete")
        if value is None:  # pragma: no cover
            return

        num_changed = self.vault.delete_metadata(key, value)
        if num_changed == 0:
            alerts.warning(f"No notes found matching: {key}: {value}")
            return

        alerts.success(
            f"Deleted value [reverse]{value}[/] from key [reverse]{key}[/] in {num_changed} notes"
        )

        return

    def noninteractive_export_csv(self, path: Path) -> None:
        """Export the vault metadata to CSV."""
        self._load_vault()
        self.vault.export_metadata(export_format="json", path=str(path))
        alerts.success(f"Exported metadata to {path}")

    def noninteractive_export_json(self, path: Path) -> None:
        """Export the vault metadata to JSON."""
        self._load_vault()
        self.vault.export_metadata(export_format="json", path=str(path))
        alerts.success(f"Exported metadata to {path}")

    def rename_key(self) -> None:
        """Rename a key in the vault."""
        original_key = self.questions.ask_existing_key(
            question="Which key would you like to rename?"
        )
        if original_key is None:  # pragma: no cover
            return

        new_key = self.questions.ask_new_key()
        if new_key is None:  # pragma: no cover
            return

        num_changed = self.vault.rename_metadata(original_key, new_key)
        if num_changed == 0:
            alerts.warning("No notes were changed")
            return

        alerts.success(
            f"Renamed [reverse]{original_key}[/] to [reverse]{new_key}[/] in {num_changed} notes"
        )

    def rename_inline_tag(self) -> None:
        """Rename an inline tag."""
        original_tag = self.questions.ask_existing_inline_tag(question="Which tag to rename?")
        if original_tag is None:  # pragma: no cover
            return

        new_tag = self.questions.ask_new_tag("New tag")
        if new_tag is None:  # pragma: no cover
            return

        num_changed = self.vault.rename_inline_tag(original_tag, new_tag)
        if num_changed == 0:
            alerts.warning("No notes were changed")
            return

        alerts.success(
            f"Renamed [reverse]{original_tag}[/] to [reverse]{new_tag}[/] in {num_changed} notes"
        )
        return

    def rename_value(self) -> None:
        """Rename a value in the vault."""
        key = self.questions.ask_existing_key(question="Which key contains the value to rename?")
        if key is None:  # pragma: no cover
            return

        question_key = Questions(vault=self.vault, key=key)
        value = question_key.ask_existing_value(question="Which value would you like to rename?")
        if value is None:  # pragma: no cover
            return

        new_value = question_key.ask_new_value()
        if new_value is None:  # pragma: no cover
            return

        num_changes = self.vault.rename_metadata(key, value, new_value)
        if num_changes == 0:
            alerts.warning("No notes were changed")
            return

        alerts.success(f"Renamed '{key}:{value}' to '{key}:{new_value}' in {num_changes} notes")

    def review_changes(self) -> None:
        """Review all changes in the vault."""
        changed_notes = self.vault.get_changed_notes()

        if len(changed_notes) == 0:
            alerts.info("No changes to review.")
            return

        alerts.info(f"Found {len(changed_notes)} changed notes in the vault")
        choices: list[dict[str, Any] | questionary.Separator] = [questionary.Separator()]
        for n, note in enumerate(changed_notes, start=1):
            _selection = {
                "name": f"{n}: {note.note_path.relative_to(self.vault.vault_path)}",
                "value": n - 1,
            }
            choices.append(_selection)

        choices.append(questionary.Separator())
        choices.append({"name": "Return", "value": "return"})

        while True:
            note_to_review = self.questions.ask_selection(
                choices=choices,
                question="Select an updated note to view the diff",
            )
            if note_to_review is None or note_to_review == "return":
                break
            changed_notes[note_to_review].print_diff()

    def transpose_metadata(self, begin: MetadataType, end: MetadataType) -> None:  # noqa: PLR0911
        """Transpose metadata from one format to another.

        Args:
            begin: The format to transpose from.
            end: The format to transpose to.
        """
        choices = [
            {"name": f"Transpose all {begin.value} to {end.value}", "value": "transpose_all"},
            {"name": "Transpose a key", "value": "transpose_key"},
            {"name": "Transpose a value", "value": "transpose_value"},
            {"name": "Back", "value": "back"},
        ]
        match self.questions.ask_selection(choices=choices, question="Select an action to perform"):
            case "transpose_all":
                num_changed = self.vault.transpose_metadata(
                    begin=begin,
                    end=end,
                    location=self.vault.insert_location,
                )

                if num_changed == 0:
                    alerts.warning("No notes were changed")
                    return

                alerts.success(f"Transposed {begin.value} to {end.value} in {num_changed} notes")
            case "transpose_key":
                key = self.questions.ask_existing_key(question="Which key to transpose?")
                if key is None:  # pragma: no cover
                    return

                num_changed = self.vault.transpose_metadata(
                    begin=begin,
                    end=end,
                    key=key,
                    location=self.vault.insert_location,
                )

                if num_changed == 0:
                    alerts.warning("No notes were changed")
                    return

                alerts.success(
                    f"Transposed key: `{key}` from {begin.value} to {end.value} in {num_changed} notes"
                )
            case "transpose_value":
                key = self.questions.ask_existing_key(question="Which key contains the value?")
                if key is None:  # pragma: no cover
                    return

                questions2 = Questions(vault=self.vault, key=key)
                value = questions2.ask_existing_value(question="Which value to transpose?")
                if value is None:  # pragma: no cover
                    return

                num_changed = self.vault.transpose_metadata(
                    begin=begin,
                    end=end,
                    key=key,
                    value=value,
                    location=self.vault.insert_location,
                )

                if num_changed == 0:
                    alerts.warning("No notes were changed")
                    return

                alerts.success(
                    f"Transposed key: `{key}:{value}` from {begin.value} to {end.value} in {num_changed} notes"
                )
            case _:
                return



"""Logging and alerts."""
import sys
from enum import Enum
from pathlib import Path
from textwrap import wrap

import rich.repr
import typer
from loguru import logger

from obsidian_metadata._utils.console import console


class LogLevel(Enum):
    """Enum for log levels."""

    TRACE = 5
    DEBUG = 10
    INFO = 20
    SUCCESS = 25
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
    EXCEPTION = 60


class VerboseLevel(Enum):
    """Enum for verbose levels."""

    WARN = 0
    INFO = 1
    DEBUG = 2
    TRACE = 3


def dryrun(msg: str) -> None:
    """Print a message if the dry run flag is set.

    Args:
        msg: Message to print
    """
    console.print(f"[cyan]DRYRUN   | {msg}[/cyan]")


def success(msg: str) -> None:
    """Print a success message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[green]SUCCESS  | {msg}[/green]")


def warning(msg: str) -> None:
    """Print a warning message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[yellow]WARNING  | {msg}[/yellow]")


def error(msg: str) -> None:
    """Print an error message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[red]ERROR    | {msg}[/red]")


def notice(msg: str) -> None:
    """Print a notice message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[bold]NOTICE   | {msg}[/bold]")


def info(msg: str) -> None:
    """Print a notice message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"INFO     | {msg}")


def usage(msg: str, width: int = 80) -> None:
    """Print a usage message without using logging.

    Args:
        msg: Message to print
        width (optional): Width of the message
    """
    for _n, line in enumerate(wrap(msg, width=width)):
        if _n == 0:
            console.print(f"[dim]USAGE    | {line}")
        else:
            console.print(f"[dim]         | {line}")


def debug(msg: str) -> None:
    """Print a debug message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[blue]DEBUG    | {msg}[/blue]")


def dim(msg: str) -> None:
    """Print a message in dimmed color.

    Args:
        msg: Message to print
    """
    console.print(f"[dim]{msg}[/dim]")


def _log_formatter(record: dict) -> str:
    """Create custom log formatter based on the log level.  This effects the logs sent to stdout/stderr but not the log file."""
    if (
        record["level"].name == "INFO"
        or record["level"].name == "SUCCESS"
        or record["level"].name == "WARNING"
    ):
        return "<level>{level: <8}</level> | <level>{message}</level>\n{exception}"

    return "<level>{level: <8}</level> | <level>{message}</level> <fg #c5c5c5>({name}:{function}:{line})</fg #c5c5c5>\n{exception}"


@rich.repr.auto
class LoggerManager:
    """Instantiate the loguru logging system with the following levels.

        - TRACE: Usage: log.trace("")
        - DEBUG: Usage: log.debug("")
        - INFO: Usage: log.info("")
        - WARNING: Usage: log.warning("")
        - ERROR: Usage: log.error("")
        - CRITICAL: Usage: log.critical("")
        - EXCEPTION: Usage: log.exception("")

    Attributes:
        log_file (Path): Path to the log file.
        verbosity (int): Verbosity level.
        log_to_file (bool): Whether to log to a file.
        log_level (int): Default log level (verbosity overrides this)

    Examples:
        Instantiate the logger:

            logging = _utils.alerts.LoggerManager(
                verbosity,
                log_to_file,
                log_file,
                log_level)
    """

    def __init__(
        self,
        log_file: Path = Path("/logs"),
        verbosity: int = 0,
        log_to_file: bool = False,
        log_level: int = 30,
    ) -> None:
        self.verbosity = verbosity
        self.log_to_file = log_to_file
        self.log_file = log_file
        self.log_level = log_level

        if self.log_file == Path("/logs") and self.log_to_file:  # pragma: no cover
            console.print("No log file specified")
            raise typer.Exit(1)

        if self.verbosity >= VerboseLevel.TRACE.value:
            logger.remove()
            logger.add(
                sys.stderr,
                level="TRACE",
                format=_log_formatter,  # type: ignore[arg-type]
                backtrace=False,
                diagnose=True,
            )
            self.log_level = 5
        elif self.verbosity == VerboseLevel.DEBUG.value:
            logger.remove()
            logger.add(
                sys.stderr,
                level="DEBUG",
                format=_log_formatter,  # type: ignore[arg-type]
                backtrace=False,
                diagnose=True,
            )
            self.log_level = 10
        elif self.verbosity == VerboseLevel.INFO.value:
            logger.remove()
            logger.add(
                sys.stderr,
                level="INFO",
                format=_log_formatter,  # type: ignore[arg-type]
                backtrace=False,
                diagnose=True,
            )
            self.log_level = 20
        else:
            logger.remove()
            logger.add(
                sys.stderr,
                format=_log_formatter,  # type: ignore[arg-type]
                level="WARNING",
                backtrace=False,
                diagnose=True,
            )
            self.log_level = 30

        if self.log_to_file is True:
            logger.add(
                self.log_file,
                rotation="5 MB",
                level=self.log_level,
                backtrace=False,
                diagnose=True,
                delay=True,
            )
            logger.debug(f"Logging to file: {self.log_file}")

        logger.debug("Logging instantiated")

    def is_trace(self, msg: str | None = None) -> bool:
        """Check if the current log level is TRACE.

        Args:
            msg (optional): Message to print. Defaults to None.

        Returns:
            bool: True if the current log level is TRACE or lower, False otherwise.
        """
        if self.log_level <= LogLevel.TRACE.value:
            if msg:
                console.print(msg)
            return True
        return False

    def is_debug(self, msg: str | None = None) -> bool:
        """Check if the current log level is DEBUG.

        Args:
            msg (optional): Message to print. Defaults to None.

        Returns:
            bool: True if the current log level is DEBUG or lower, False otherwise.
        """
        if self.log_level <= LogLevel.DEBUG.value:
            if msg:
                console.print(msg)
            return True
        return False

    def is_info(self, msg: str | None = None) -> bool:
        """Check if the current log level is INFO.

        Args:
            msg (optional): Message to print. Defaults to None.

        Returns:
            bool: True if the current log level is INFO or lower, False otherwise.
        """
        if self.log_level <= LogLevel.INFO.value:
            if msg:
                console.print(msg)
            return True
        return False

    def is_default(self, msg: str | None = None) -> bool:
        """Check if the current log level is default level (SUCCESS or WARNING).

        Args:
            msg (optional): Message to print. Defaults to None.

        Returns:
            bool: True if the current log level is default or lower, False otherwise.
        """
        if self.log_level <= LogLevel.WARNING.value:
            if msg:
                console.print(msg)
            return True
        return False  # pragma: no cover
